use std::{
    borrow::Cow,
    ffi::{OsStr, OsString},
    fmt::Display,
    ops::{Deref, DerefMut},
    path::PathBuf,
};

use derive_more::Display;
use osstr_traits::{OsDisplay, ToOsString};

type OsStrDisplay<'a> = std::ffi::os_str::Display<'a>;
type PathDisplay<'a> = std::path::Display<'a>;

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct DelimitedVec<T, const D: char = ' '>(Vec<T>);

impl<T: OsDisplay, const D: char> OsDisplay for DelimitedVec<T, D> {
    fn fmt_os(&self, f: &mut osstr_traits::OsStringFormatter) -> std::fmt::Result {
        let mut first = true;

        for item in &self.0 {
            if !first {
                f.write_os_str(&D.to_os_string())?;
            }

            item.fmt_os(f)?;
            first = false;
        }

        Ok(())
    }
}

impl<T: Display, const D: char> Display for DelimitedVec<T, D> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut first = true;

        for item in &self.0 {
            if !first {
                write!(f, "{D}")?;
            }

            write!(f, "{item}")?;
            first = false;
        }

        Ok(())
    }
}

impl<A, const D: char> FromIterator<A> for DelimitedVec<A, D> {
    fn from_iter<T: IntoIterator<Item = A>>(iter: T) -> Self {
        Self(iter.into_iter().collect())
    }
}

impl<'a, const D: char> From<&'a Vec<PathBuf>> for DelimitedVec<PathDisplay<'a>, D> {
    fn from(value: &'a Vec<PathBuf>) -> Self {
        let iter = value.iter().map(|s| s.display());
        Self::from_iter(iter)
    }
}

impl<'a, const D: char> From<&'a Vec<OsString>> for DelimitedVec<OsStrDisplay<'a>, D> {
    fn from(value: &'a Vec<OsString>) -> Self {
        let iter = value.iter().map(|s| s.display());
        Self::from_iter(iter)
    }
}

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct CommandArgs<T>(pub Vec<T>);

impl<T: OsDisplay> OsDisplay for CommandArgs<T> {
    fn fmt_os(&self, f: &mut osstr_traits::OsStringFormatter) -> std::fmt::Result {
        let mut first = true;

        for item in &self.0 {
            if !first {
                f.write_os_str(OsStr::new(" "))?;
            }

            item.fmt_os(f)?;
            first = false;
        }

        Ok(())
    }
}

impl<T: Display> Display for CommandArgs<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut first = true;

        for item in &self.0 {
            if !first {
                write!(f, " ")?;
            }

            write!(f, "{item}")?;
            first = false;
        }

        Ok(())
    }
}

impl<T> Default for CommandArgs<T> {
    fn default() -> Self {
        Self(Vec::new())
    }
}

impl<T> Deref for CommandArgs<T> {
    type Target = Vec<T>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl<T> DerefMut for CommandArgs<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

impl<A> FromIterator<A> for CommandArgs<A> {
    fn from_iter<T: IntoIterator<Item = A>>(iter: T) -> Self {
        Self(iter.into_iter().collect())
    }
}

impl<T> IntoIterator for CommandArgs<T> {
    type Item = T;
    type IntoIter = <Vec<T> as IntoIterator>::IntoIter;

    fn into_iter(self) -> Self::IntoIter {
        self.0.into_iter()
    }
}

#[derive(Clone, Debug, Display)]
#[display("{} {}", command.display(), args)]
pub struct Command<'a, T> {
    command: Cow<'a, OsStr>,
    args: CommandArgs<T>,
}

impl<'a, T> Command<'a, T> {
    pub fn new(command: impl Into<Cow<'a, OsStr>>) -> Self {
        Self {
            command: command.into(),
            args: CommandArgs::default(),
        }
    }

    pub fn arg(&mut self, arg: T) {
        self.args.push(arg);
    }
}

impl<'a, T: AsRef<OsStr>> Command<'a, T> {
    pub fn into_os_command(self) -> std::process::Command {
        let mut cmd = std::process::Command::new(self.command);
        cmd.args(self.args);
        cmd
    }
}
