use std::{
    borrow::Cow,
    env,
    path::{Path, PathBuf},
};

use confique::Config;
use directories::ProjectDirs;
use serde::{Deserialize, Deserializer, Serialize};

use crate::error::{Error, ShellExpansionError};

trait Fallback: Config {
    fn fallback() -> <Self as Config>::Partial;
}

fn project_dir() -> Option<ProjectDirs> {
    ProjectDirs::from("cafe", "kitsu", "TormentNexus")
}

fn expand_tilde<'a>(path: impl Into<Cow<'a, Path>>) -> Result<Cow<'a, Path>, ShellExpansionError> {
    let path = path.into();
    let into_err = || ShellExpansionError::new("~", path.to_str());

    if !path.starts_with("~") {
        Ok(path)
    } else if path == Path::new("~") {
        env::home_dir().map(Cow::Owned).ok_or_else(into_err)
    } else {
        let mut home_dir = env::home_dir().ok_or_else(into_err)?;
        if home_dir == Path::new("/") {
            let stripped = path.strip_prefix("~").map_err(|_e| into_err())?;
            Ok(Cow::Owned(stripped.into()))
        } else {
            home_dir.push(path.strip_prefix("~/").map_err(|_e| into_err())?);
            Ok(Cow::Owned(home_dir))
        }
    }
}

fn deserialize_path<'de, D: Deserializer<'de>>(de: D) -> Result<PathBuf, D::Error> {
    let path = PathBuf::deserialize(de)?;
    if let Ok(expanded_path) = expand_tilde(&path) {
        Ok(expanded_path.into())
    } else {
        Ok(path)
    }
}

#[derive(Config, Debug, Serialize, Deserialize)]
pub struct EditorCfg {
    #[config(deserialize_with = deserialize_path, env = "EDITOR_PATH")]
    pub path: PathBuf,
}

impl Fallback for EditorCfg {
    fn fallback() -> <Self as Config>::Partial {
        Self::Partial {
            path: project_dir().map(|d| d.data_local_dir().join("editors/")),
        }
    }
}

#[derive(Debug, Config, Serialize, Deserialize)]
pub struct ProjectCfg {
    #[config(deserialize_with = deserialize_path, env = "PROJECT_MANIFEST_PATH")]
    pub manifest_path: PathBuf,
}

impl Fallback for ProjectCfg {
    fn fallback() -> <Self as Config>::Partial {
        Self::Partial {
            manifest_path: project_dir().map(|d| d.data_local_dir().join("manifest.toml")),
        }
    }
}
#[derive(Debug, Config, Serialize, Deserialize)]
pub struct Cfg {
    #[config(nested)]
    pub editor: EditorCfg,

    #[config(nested)]
    pub project: ProjectCfg,
}

impl Cfg {
    pub fn from_env() -> Result<Cfg, Error> {
        let mut builder = Cfg::builder().env();

        if let Some(dir) = project_dir() {
            builder = builder.file(dir.config_dir().join("config.toml"));
        }

        builder
            .preloaded(Self::fallback())
            .load()
            .map_err(Into::into)
    }
}

impl Fallback for Cfg {
    fn fallback() -> <Self as Config>::Partial {
        Self::Partial {
            editor: EditorCfg::fallback(),
            project: ProjectCfg::fallback(),
        }
    }
}
