#![feature(min_specialization)]

pub mod collection;
pub mod config;
pub mod editor;
pub mod editor_command;
pub mod error;
pub mod project;

use std::{fs, ops::Deref, path::Path};

use serde::{Deserialize, Serialize};

use crate::{
    config::Cfg,
    editor::Editor,
    error::{Error, FileError, ProjectManifestError},
    project::Project,
};

#[derive(Debug, Serialize, Deserialize)]
pub struct Projects {
    projects: Vec<Project>,
}

impl Projects {
    pub fn new(projects: Vec<Project>) -> Self {
        Self { projects }
    }
}

impl From<Vec<Project>> for Projects {
    fn from(value: Vec<Project>) -> Self {
        Self::new(value)
    }
}

impl Default for Projects {
    fn default() -> Self {
        Self::new(Vec::new())
    }
}

impl Deref for Projects {
    type Target = Vec<Project>;

    fn deref(&self) -> &Self::Target {
        &self.projects
    }
}

#[derive(Debug)]
pub struct TormentNexus {
    config: Cfg,
    projects: Projects,
}

impl TormentNexus {
    pub fn new(config: Cfg, projects: impl Into<Projects>) -> Self {
        Self {
            config,
            projects: projects.into(),
        }
    }

    pub fn load() -> Result<Self, Error> {
        let config = Cfg::from_env()?;
        let projects = Self::read_projects(&config.project.manifest_path)?;

        Ok(Self::new(config, projects))
    }

    pub fn read_projects(
        manifest_path: impl AsRef<Path>,
    ) -> Result<Projects, ProjectManifestError> {
        let content = fs::read_to_string(&manifest_path)
            .map_err(|err| FileError::new(manifest_path.as_ref(), err))?;

        Ok(toml::from_str::<Projects>(&content)?)
    }

    pub fn installed_editors(&self) -> Vec<Editor> {
        Vec::new()
    }

    pub fn projects(&self) -> &Projects {
        &self.projects
    }
}

#[cfg(test)]
mod tests {
    use crate::TormentNexus;

    #[test]
    fn without_cfg() {
        let hub = TormentNexus::load();
        println!("{hub:?}");
        assert!(hub.is_ok())
    }
}
